#include "bootmaster.h"

///////////////////////////////////////////////////////////////////////////////
//   Macros

#define screen_width	512
#define screen_height	480

#define	GFXPIPE_SIZE	0x2000

///////////////////////////////////////////////////////////////////////////////
//   Externs

///////////////////////////////////////////////////////////////////////////////
//   Globals

SJGFX_screenEnv screenEnv;
char padBuf1[256] __attribute__((aligned(64)));
unsigned char gfxPipeMem[GFXPIPE_SIZE];
char *bootFileName;

///////////////////////////////////////////////////////////////////////////////
//   Function declarations

static void initGFX();
int loadModules();
int initPad(int port, int slot, char* buffer);
int getInput();
char *parseSystemCnf();
void waitCdReady();
static void printMsg(char *string);
static void error(int errno);

///////////////////////////////////////////////////////////////////////////////
//   Function defenitions

int main()
{
	int retVal;

	SJGFX_Disable();

	// Initialise
	SifInitRpc(0);
	SifResetIop();
	SifInitRpc(0);

	// Initialise graphics subsystem
	initGFX();

	// Load required IRX modules
	loadModules();
	
	// Attempt to initialise the EE cdvdfsv RPC server
	if(!CdInit(CdInitNoChk)) error(ERROR_CDVDFSV_INIT);

	// Initialise the joypad subsystem
	padInit(0);
	initPad(0, 0, padBuf1);

	// Enable the display
	SJGFX_ClearScr();
	SJGFX_Enable();

	if(initHax0r() < 0)
		error(ERROR_HAX0R);
	
	// Spin down the cd
	CdStop();

	printMsg("Swap CD/DVD's now\n\nPress "DRAW_X" to start game normally\nPress "DRAW_O" to force NTSC\nPress "DRAW_TRI" to force PAL");

	// Wait for X or O to be pressed
	while(!(retVal = getInput()));

	// If O or Tri was pressed, install patcher function for sceGsSetCrt
	if(retVal == 2) installPNSelector(2);
	else if(retVal == 3) installPNSelector(3);

	printMsg("Loading and parsing SYSTEM.CNF");

	// Wait until the CD has spun up and in a ready state
	waitCdReady();

	// Attempt to parse SYSTEM.CNF in order to get the filename of the game elf
	if((int)(bootFileName = parseSystemCnf()) <= 0)
	{
		error((int)bootFileName);
		SleepThread();
	}

	printMsg("Loading elf");

	// Shutdown
	SJGFX_Cleanup();
	padPortClose(0, 0);
	SifExitRpc();

	LoadExecPS2((const char *)bootFileName, 0, NULL);

	// Should never get here
	SleepThread();
}


static void initGFX()
{
	// Detect video mode, set display parameters accordingly
	if(pal_ntsc() == 3)
	{
		screenEnv.videoMode = 0; // pal
		screenEnv.screenX = 130;
		screenEnv.screenY = 90;
	} else {
		screenEnv.videoMode = 1; // ntsc
		screenEnv.screenX = 130;
		screenEnv.screenY = 45;
	}

	screenEnv.screenWidth = screen_width;
	screenEnv.screenHeight = screen_height;
	screenEnv.interlace = GS_INTERLACE;
	screenEnv.frameBufPSM = GS_PSMCT32;
	screenEnv.zBufPSM = GS_PSMZ16S;
	screenEnv.gfxPipeMem = gfxPipeMem;
	screenEnv.gfxPipeSize = GFXPIPE_SIZE;

	// Initialise the drawing and display environment
	SJGFX_init(&screenEnv);

	// Setup the default font
//	defaultPrintStyle.size = 2.0f;

	// The following code had to be added, or GFX wouldnt work after IOP reboot?!?!
	SJGFX_finish();
	SJGFX_vsync();
	SJGFX_switchFrameBuffers();
}

int loadModules()
{
	SifLoadModule("rom0:SIO2MAN", 0, NULL);
	SifLoadModule("rom0:PADMAN", 0, NULL);

	return 0;
}

int initPad(int port, int slot, char* buffer)
{
    int ret;
	int timeout = 0;

	// Open pad port
	if((ret = padPortOpen(port, slot, buffer)) == 0) {
		return -1;
    }

	// Try to wait for the pad to became stable
	ret = 0;
	while(!(ret & PAD_STATE_FINDCTP1) || !(ret & PAD_STATE_STABLE)) {
		ret = padGetState(port, slot);
		if(++timeout > 50) break;
		WaitForNextVRstart(1);
	}

	// Enable analog mode
	if(padGetState(port, slot) == PAD_STATE_STABLE)
		padSetMainMode(port, slot, PAD_MMODE_DUALSHOCK, PAD_MMODE_UNLOCK);

    return 0;
}

int getInput()
{
	static int old_pad = 0;
	int new_pad = 0;
	int pad1_data = 0;
	static struct padButtonStatus pad1;
	static int pad1_connected = 0;
	int state_ret;

	// Check if pad is connected
	state_ret = padGetState(0, 0);
	if((state_ret == PAD_STATE_STABLE) || (state_ret == PAD_STATE_FINDCTP1)) {
		if(pad1_connected == 0) {
			WaitForNextVRstart(1);
		}
		pad1_connected = 1;
	} else pad1_connected = 0;

	// Update controller status
	if(pad1_connected) {
		padRead(0, 0, &pad1); // port, slot, buttons
		pad1_data = 0xffff ^ ((pad1.btns[0] << 8) | pad1.btns[1]);

		if((pad1.mode >> 4) == 0x07) {
			if(pad1.ljoy_v < 64) pad1_data |= PAD_UP;
			else if(pad1.ljoy_v > 192) pad1_data |= PAD_DOWN;

			if(pad1.ljoy_h < 64) pad1_data |= PAD_LEFT;
			else if(pad1.ljoy_h > 192) pad1_data |= PAD_RIGHT;
		}
	}

	new_pad = pad1_data & ~old_pad;
	old_pad = pad1_data;

	if(new_pad & PAD_CROSS) return 1;
	if(new_pad & PAD_CIRCLE) return 2;
	if(new_pad & PAD_TRIANGLE) return 3;

	return 0;
}

// Parses system.cnf. Returns the boot elf filename on success, NULL on failure
char *parseSystemCnf()
{
	char *buffer;
	int fd, fdSize;
	int i;

	// Open SYSTEM.CNF on the cdrom, allocate memory for it, terminate the array
	fd = fioOpen("cdrom0:\\SYSTEM.CNF;1", O_RDONLY);
	if(fd < 0) return (char *)ERROR_SYSTEMCNF_FILEIO;

	fdSize = fioLseek(fd, 0, SEEK_END);
	fioLseek(fd, 0, SEEK_SET);

	buffer = malloc(fdSize + 1);
	if(!buffer) return (char *)ERROR_SYSTEMCNF_MEMORY;

	if(fioRead(fd, buffer, fdSize) != fdSize) return (char *)ERROR_SYSTEMCNF_FILEIO;
	fioClose(fd);
	buffer[fdSize] = '\0';

	// Find boot file substring
	buffer = strstr(buffer, "BOOT2");
	buffer += 5;
	while((*buffer == ' ') || (*buffer == '\t')) buffer++;
	buffer++; // bypass '=' character
	while((*buffer == ' ') || (*buffer == '\t')) buffer++;

	i = 0;
	while((buffer[i] != '\n') && (buffer[i] != '\r')) i++;

	// Terminate string at end of boot elf filename
	buffer[i] = '\0';

	// Return pointer to boot elf filename string
	return buffer;
}

void waitCdReady()
{
	// Block until the dvdrom is ready to take commands
	CdDiskReady(0);

	// Spin up the cd/dvd
	CdStandby();
}

static void printMsg(char *string)
{
	int i;

	for(i = 0; i < 2; i++)
	{
		SJGFX_print(0, screen_width, screen_height / 2 - 32, 2, GFX_FONT_ALIGN_CENTER, &defaultPrintStyle, string);

		SJGFX_finish();
		SJGFX_vsync();
		SJGFX_switchFrameBuffers();
	}
}

static void error(int errno)
{
	SJGFX_ClearScr();
	SJGFX_Enable();

	while(1)
	{
		SJGFX_print(0, screen_width, screen_height / 2 - 32, 2, GFX_FONT_ALIGN_CENTER, &defaultPrintStyle, "FATAL ERROR! Code: %d", errno * -1);

		SJGFX_finish();
		SJGFX_vsync();
		SJGFX_switchFrameBuffers();
	}
}
